#!/usr/bin/env python
# coding: utf-8

import numpy as np
import gc
import math
import time
import json


# function that computes the exact degree of the polynomial given the type and bs
def poldeg(seq_type,qs,bs):
    deg = 0
    occ = np.zeros(k, dtype=int) # occurences of the variables x_i, i=1,...,k in all the polynomial factors
    for i in range(0, k-1):
        for j in range(i+1, k):
            if qs[j] == qs[i]:
                occ[i] = occ[i] + 1
                occ[j] = occ[j] + 1
                deg = deg + 1
            if bs[j+1] == bs[i] and (i != 0 or j != k-1):
                for l in range(i, j+1):
                    occ[l] = occ[l] + 1
                deg = deg + 1
    return deg, occ

# function that removes the unnecessary terms
def remTerms(f, ubound, occs):
    R = f.parent()
    d = f.dict()
    k = len(ubound)
    dd = {ee: c for ee, c in d.items() if all(ee[i] <= ubound[i] and ee[i] >= ubound[i]-occs[i] for i in range(0, k))}
    return R(dd)

# Open a json file containing all the types and associated cosets for a given k

path = "json/K10Z2.json" # json file containing all the types and associated monomials
fileR = open(path)
dataset = json.load(fileR)
fileR.close()

# Defining the initial parameters and polynomial ring

k = sum([int(x) for x in next(iter(dataset.keys())).strip(')][(').split(',')]) # extracting the k value
R = PolynomialRing(ZZ,k,"x")
vars = R.gens()
print("Variables: ", vars, "\n")

allCorrect = True

for typ, coset in dataset.items():
    
    seq_type = [int(x) for x in typ.strip(')][(').split(',')]
    
    t = len(seq_type)
    
    print("Type: ", seq_type, "\n")
    
    for qseq, monomials in coset.items():
        qs = [int(x) for x in qseq.strip(')][(').split(',')]
        seq_test = [qs.count(i) for i in range(0, t)]
        
        bs = np.mod(np.cumsum(qs), t) # partial sum quotient sequence
        bs = np.insert(bs, 0, 0) # first element in the partial sum must be 0
        
        pdeg, occs = poldeg(seq_type, qs, bs)
        print("Quotient sequence: ", qs)
        print("Cumsum quotient sequence: ", bs)
        print("Polynomial degree: ", pdeg)
        print("Variables occurrence: ", occs)
        
        for monomial, coeff in monomials.items():
            
            coccs = occs.copy()
        
            bounding_exponents = [int(x) for x in monomial.strip(')][(').split(',')]
            
            print("Exponents of the monomial: ", bounding_exponents)

            check_degree = True
            for i in range(0, k):
                if bounding_exponents[i] > seq_type[qs[i]]-1:
                    check_degree = False
                    break

            p = 1
            nitems = 0
            for i in range(0, k-1):
                for j in range(i+1, k):
                    remove = False
                    f1 = 1
                    f2 = 1
                    if qs[j] == qs[i]:
                        f1 = (vars[j] - vars[i])
                        coccs[i] = coccs[i] - 1
                        coccs[j] = coccs[j] - 1
                        remove = True
                        nitems = nitems + 1
                    if bs[j+1] == bs[i] and (i != 0 or j != k-1):
                        f2 = sum(vars[k] for k in range(i, j+1))
                        for l in range(i, j+1):
                            coccs[l] = coccs[l] - 1
                        remove = True
                        nitems = nitems + 1
                    ff = f1 * f2
                    p = p * ff
                    if remove == True:
                        # we remove the terms that do not respect the degree of the targeting monomial
                        p = remTerms(p, bounding_exponents, coccs)
            print("Monomial: ", p)
            print("Coefficient: ", factor(p.lc()))
            testcorr = (p.total_degree() == pdeg and seq_type==seq_test and check_degree)
            print("Is correct? ", testcorr, "\n")

            allCorrect = allCorrect and testcorr

print("All the tests are correct? ", allCorrect)